from PySide6 import QtCore
from obspy import Trace

# Global flag that determines whether the user requested to stop processing and
# its corresponding global lock.
stop = False
stop_lock = QtCore.QMutex()


class DecimatorWorkerSignals(QtCore.QObject):
    """
    Object that contains the signals for DownsamplerWorker.

    Signals:
        finished: emitted when a DownsamplerWorker instance finishes, contains
            the result of the computation done by the instance
        stopped: emitted when a DownsamplerWorker instance is stopped as a
            result of a request from the user
    """
    finished = QtCore.Signal(Trace)
    stopped = QtCore.Signal()


class DecimatorWorker(QtCore.QRunnable):
    """
    The worker class that executes the code to downsample data for plotting.
    """

    def __init__(self, trace, decimate_factor, do_decimate=True):
        super().__init__()

        self.trace = trace
        self.decimate_factor = decimate_factor

        self.signals = DecimatorWorkerSignals()

        # Whether to downsample the data. Intended for times when a user of
        # this class has some data sets that might not need downsampling.
        self.do_decimate = do_decimate

    def run(self):
        global stop, stop_lock
        try:
            # stop_locker = QMutexLocker(stop_lock)
            # stop_lock.lock()
            if stop:
                # stop_lock.unlock()
                raise StopRequested
            # stop_lock.unlock()
            elif self.do_decimate:
                self.trace.decimate(self.decimate_factor)
        except StopRequested:
            # The way we set things up, only one background thread can do
            # downsampling at a time. When we stop processing waveform data,
            # all other queued background downsampling tasks are killed,
            # leaving the sole running downsampler. As a result, we can reset
            # the stop flag with no issue because no other downsampler will
            # read it afterward.
            stop = False
            self.signals.stopped.emit()
        else:
            if not stop:
                self.signals.finished.emit(
                    self.trace
                )


class Decimator:
    """
    The class that coordinate the downsampling of data in background threads.
    Intended to be used in cases where there are multiple data sets that need
    to be downsampled. For instances where there is only one data set,
    DownsamplerWorker offers an easier-to-use API at the cost of having to
    do your own thread management.
    """

    def __init__(self):
        self.thread_pool = QtCore.QThreadPool()
        # Setting the maximum number of thread executed in the background in
        # this thread pool to 1. Doing so helps to prevent the situation where
        # multiple threads finish their work at once, which may lead to the
        # main thread hanging if the post-processing procedure is too complex.
        self.thread_pool.setMaxThreadCount(1)

        self.worker_list = []

    def add_worker(self, trace, decimate_factor, do_decimate=True):
        """
        Create a downsampler worker for a data set and add it to the worker
        list. Also returns the created worker so that users can connect its
        signals to their slots.

        :param times: the times array to downsample
        :param data: the data array to downsample
        :param log_indexes: the soh message line indices array to downsample
        :param rq_points: the requested size of the downsampled arrays
        :param do_decimate: whether to downsample the given data set. True by
            default
        :return: the worker that will downsample the given data
        """
        worker = DecimatorWorker(trace, decimate_factor, do_decimate)
        self.worker_list.append(worker)

        return worker

    def start(self):
        """
        Signal the internal thread pool to put all the stored downsampler
        workers into its execution queue. Each worker will begin running when
        its turn comes.
        """
        for worker in self.worker_list:
            self.thread_pool.start(worker)

    def request_stop(self):
        """
        Request the downsampler to stop all its workers by setting the global
        stop flag.
        """
        global stop, stop_lock
        stop_lock.lock()
        stop = True
        stop_lock.unlock()
        self.thread_pool.clear()


class StopRequested(Exception):
    """
    Exception that is raised to indicate that a downsampler worker should stop
    running and starts cleaning up.
    """
    pass
